/* utilities.c  - a set of functions that define a simplified interface to EXS.
 *		   These functions encapsulate all the steps needed to:
 *			- read and write blocks of a specified size
 *			- open clients and servers
 *		   Modified for exs API.
 * on Solaris this program must be linked with -lnsl and -lsocket
 */

/*
 * Copyright 2007, 2008, 2009, 2010, 2011, 2012 by
 *	University of New Hampshire Interoperability Laboratory
 *	121 Technology Drive, Suite 2
 *	Durham, NH 03824-4716
 *	phone: 1-603-862-0090
 *	fax: 1-603-862-4181
 *	e-mail: info@iol.unh.edu
 * All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grant No. OCI-1127228.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

/*
 * The UNH IOL implementation of the Extended Sockets API (ES-API) Issue 1.0
 *
 * defined in:	The Interconnect Software Consortium Issue 1.0
 *		ISBN: 1-931624-52-6
 *		Published by The open Group, January 2005.
 *
 * all page references in this source code are to pages in that document.
 */

/*
 * Most of the coding in this file done by:
 *	Bob Russell
 *	rdr@iol.unh.edu
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define _POSIX_C_SOURCE 200112L
#define _ISOC99_SOURCE
#define __EXTENSIONS__
#define _XOPEN_SOURCE 600

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <netdb.h>

#include "utilities.h"

#ifdef ALLOW_DEBUG
unsigned int util_debug_mask;
#endif

#define CONNECT_DEPTH 1
#define CONNECT_ITEMS 1
#define SEND_ITEMS 1
#define RECV_ITEMS 1

#define TEXT_SIZE	127

static int	family[2] = {AF_INET, AF_INET6};


static int utility_debug_flag = 0;

void
blockio_set_debug(int flag)
{
	fprintf(stderr, "set utility_debug_flag from %d to %d\n",
		utility_debug_flag, flag);
	utility_debug_flag = flag;
}	/* blockio_set_debug */


/* Returns after starting write of nbytes characters from buffer to fd.  */
int
asynch_send(int fd, char *buffer, int nbytes, int flags,
		exs_mhandle_t buf_handle, exs_qhandle_t wait_queue)
{
	TRACE_DBG(LOG_UTILITIES,
		"asynch_send calling exs_send, fd %d, buffer %p, len %d, wait_queue %p, buf_handle %p",
		fd, buffer, nbytes, wait_queue, buf_handle);
		
	if (exs_send(fd,buffer,nbytes,flags,wait_queue,NULL,buf_handle) < 0) {
		perror("asynch_send exs_send");
		return -1;
	}

	TRACE_DBG(LOG_UTILITIES,
		"asynch_send back from exs_send, fd %d, buffer %p, len %d, wait_queue %p, buf_handle %p",
		fd, buffer, nbytes, wait_queue, buf_handle);

	return nbytes;
}	/* asynch_send */


/* Returns after starting write of count vectors from sendvec to fd.  */
int
asynch_sendfile(int fd, exs_xferfile_t *sendvec, int count,
		exs_qhandle_t wait_queue, exs_ahandle_t ahandle)
{
	TRACE_DBG(LOG_UTILITIES,
		"asynch_sendfile calling exs_sendfile, fd %d, sendvec %p, count %d, wait_queue %p, ahandle %p",
		fd, sendvec, count, wait_queue, ahandle);
		
	if (exs_sendfile(fd, sendvec, count, 0, wait_queue, ahandle) < 0) {
		perror("asynch_sendfile exs_sendfile");
		return -1;
	}

	TRACE_DBG(LOG_UTILITIES,
		"asynch_sendfile back from exs_sendfile, fd %d, sendvec %p, count %d, wait_queue %p, ahandle %p",
		fd, sendvec, count, wait_queue, ahandle);

	return count;
}	/* asynch_sendfile */


/* Returns after starting read of nbytes characters into buffer from fd.  */
int
asynch_recv(int fd, char *buffer, int nbytes, int flags,
		exs_qhandle_t wait_queue,
		exs_ahandle_t ahandle,
		exs_mhandle_t buf_handle)
{

	TRACE_DBG(LOG_UTILITIES,
		"asynch_recv calling exs_recv, fd %d, buffer %p, bytes %d, wait_queue %p",
		fd, buffer, nbytes, wait_queue);

	if (exs_recv(fd,buffer,nbytes,flags,wait_queue,ahandle,buf_handle) < 0){
		perror("asynch_recv exs_recv");
		return -1;
	}

	return nbytes;
}	/* asynch_recv */


/* Returns after getting an async event from wait_queue	*/
int
asynch_wait(exs_qhandle_t wait_queue, exs_event_t *vector,
		struct timeval *timeout)
{
	int nevents;

	/* in exs need to wait for asynchronous send or recv to complete */
	TRACE_DBG(LOG_UTILITIES,
		"asynch_wait calling exs_qdequeue, wait_queue %p, vector %p",
		wait_queue, vector);

	nevents = exs_qdequeue(wait_queue, vector, 1, timeout);
	if (nevents <= 0) {
		if (nevents == 0)
			errno = ETIME;	/* timer expired */
		perror("asynch_wait exs_qdequeue");
		return -1;
	}
#if defined(ALLOW_PARANOID_CHECKING)
	if (nevents != 1) {
		fprintf(stderr, "asynch_wait exs_qdequeue dequeued %d "
			"items, expected 1\n", nevents);
		return -1;
	}
#endif
	TRACE_DBG(LOG_UTILITIES,
		"asynch_wait back from exs_qdequeue, vector %p type is %d, errno %d, fd %d",
		vector, vector->exs_evt_type, vector->exs_evt_errno,
		vector->exs_evt_socket);
	if (vector->exs_evt_errno != 0) {
		errno = vector->exs_evt_errno;
		return -1;
	}

	return 0;		/* return 0 if all ok */
}	/* asynch_wait */


/* Returns after all nbytes characters from buffer have been written to fd.  */
int
writeblock(int fd, char *buffer, int nbytes, int flags,
		exs_mhandle_t buf_handle, exs_qhandle_t wait_queue)
{
	int n;			/* number of bytes actually sent in one write */
	int offset;		/* total number of bytes sent so far */
	int nevents;
	exs_event_t send_vector[SEND_ITEMS];

	for (offset = 0; offset < nbytes; offset += n) {

		TRACE_DBG(LOG_UTILITIES,
			"writeblock calling exs_send, fd %d, address %p, len %d, wait_queue %p, buf_handle %p",
			fd, &buffer[offset],
			nbytes-offset, wait_queue, buf_handle);
			
		if (exs_send(fd, &buffer[offset], nbytes-offset, flags,
				wait_queue, NULL, buf_handle)<0) {
			perror("writeblock exs_send");
			break;
		}
		/* in exs need to wait for asynchronous send to complete */
		TRACE_DBG(LOG_UTILITIES,
			"writeblock back from exs_send ok, calling exs_qdequeue, wait_queue %p, send_vector %p",
			wait_queue, send_vector);

		nevents = exs_qdequeue(wait_queue,send_vector,SEND_ITEMS,NULL);
		if (nevents < 0) {
			perror("writeblock exs_qdequeue");
			break;
		}
#if defined(ALLOW_PARANOID_CHECKING)
		if (nevents != 1) {
			fprintf(stderr, "writeblock exs_qdequeue: dequeued %d "
				"items, expected 1\n", nevents);
			break;
		}
#endif
		TRACE_DBG(LOG_UTILITIES,
			"writeblock back from exs_qdequeue, nevents = %d, send_vector = %p",
			nevents, send_vector);
#if defined(ALLOW_PARANOID_CHECKING)
		if (send_vector[0].exs_evt_type != EXS_EVT_SEND) {
			fprintf(stderr, "writeblock exs_qdequeue: dequeued item"
				" type %d, expected %d\n",
				send_vector[0].exs_evt_type,
				EXS_EVT_SEND);
			break;
		}
#endif
		if (send_vector[0].exs_evt_errno != 0) {
			errno = send_vector[0].exs_evt_errno;
			perror("writeblock exs_qdequeue status");
			break;
		}
		n = send_vector[0].exs_evt_union.exs_evt_xfer.exs_evt_length;
		TRACE_DBG(LOG_UTILITIES,
			"writeblock back from send_vector %p, n = %d",
			send_vector, n);
	}
	TRACE_DBG(LOG_UTILITIES, "writeblock returning %d", offset);

	return offset;		/* return total number of bytes actually sent */
}	/* writeblock */


/* Returns after all nbytes characters have been read into buffer from fd.
 * Will return sooner if an end of file is encountered.
 * Value returned is actual number of bytes read.
 */
int readblock(int fd, char *buffer, int nbytes, exs_qhandle_t wait_queue)
{
	int n;		/* number of bytes actually received in one read */
	int offset;	/* total number of bytes received so far */
	int nevents;
	exs_event_t recv_vector[RECV_ITEMS];

	for (offset = 0; offset < nbytes; offset += n) {
		TRACE_DBG(LOG_UTILITIES,
			"readblock calling exs_recv, fd %d, buffer %p, bytes %d, offset %d, wait_queue %p",
			fd, &buffer[offset], nbytes-offset, offset,
			wait_queue);
		if (exs_recv(fd, &buffer[offset], nbytes-offset, 0,
				wait_queue, NULL, EXS_MHANDLE_UNREGISTERED)<0){
			perror("readblock exs_recv");
			break;
		}
		/* in exs need to wait for asynchronous send to complete */
		TRACE_DBG(LOG_UTILITIES,
			"readblock calling exs_qdequeue, wait_queue %p, recv_vector %p, n_items %d",
			wait_queue, recv_vector, RECV_ITEMS);
		nevents = exs_qdequeue(wait_queue,recv_vector,RECV_ITEMS,NULL);
		TRACE_DBG(LOG_UTILITIES,
			"readblock exs_qdequeue returned nevents %d, recv_vector %p",
			nevents, recv_vector);
		if (nevents < 0) {
			perror("readblock exs_qdequeue");
			break;
		}
#if defined(ALLOW_PARANOID_CHECKING)
		if (nevents != 1) {
			fprintf(stderr, "readblock exs_qdequeue: dequeued %d "
				"items, expected 1\n", nevents);
			break;
		}
		if (recv_vector[0].exs_evt_type != EXS_EVT_RECV) {
			fprintf(stderr, "readblock exs_qdequeue: dequeued item"
				" type %d, expected %d\n",
				recv_vector[0].exs_evt_type,
				EXS_EVT_RECV);
			break;
		}
#endif
		TRACE_DBG(LOG_UTILITIES,
			"readblock exs_qdequeue item type %d is RECV, errno %d",
			recv_vector[0].exs_evt_type,
			recv_vector[0].exs_evt_errno);
		if (recv_vector[0].exs_evt_errno != 0) {
			errno = recv_vector[0].exs_evt_errno;
			perror("readblock exs_qdequeue status");
			break;
		}
		n = recv_vector[0].exs_evt_union.exs_evt_xfer.exs_evt_length;
		TRACE_DBG(LOG_UTILITIES,
			"readblock exs_qdequeue item len %d", n);
		if (n == 0) {
			/* hit EOF */
			break;
		}
	}
	return offset;		/* return total number of bytes actually read */
}	/* readblock */


/* Returns after <= nbytes characters have been read into buffer from fd.
 * Will also return if an end of file is encountered.
 * Value returned is actual number of bytes read.
 */
int readpacket(int fd, char *buffer, int nbytes, exs_mhandle_t buf_handle,
		exs_qhandle_t wait_queue)
{
	int n = 0;	/* number of bytes actually received in one read */
	int nevents;
	exs_event_t recv_vector[RECV_ITEMS];

	TRACE_DBG(LOG_UTILITIES,
		"readpacket calling exs_recv, fd %d, buffer %p, bytes %d, wait_queue %p",
		fd, buffer, nbytes, wait_queue);

	do {
		if (exs_recv(fd, buffer, nbytes, 0,
				wait_queue, NULL, buf_handle)<0){
			perror("readpacket exs_recv");
			break;
		}
		/* in exs need to wait for asynchronous send to complete */
		TRACE_DBG(LOG_UTILITIES,
			"readpacket calling exs_qdequeue, wait_queue %p, recv_vector %p, n_items %d",
			wait_queue, recv_vector, RECV_ITEMS);
		nevents = exs_qdequeue(wait_queue,recv_vector,RECV_ITEMS,NULL);
		TRACE_DBG(LOG_UTILITIES,
				"readpacket exs_qdequeue returned nevents %d, recv_vector %p",
				nevents, recv_vector);
		if (nevents < 0) {
			perror("readpacket exs_qdequeue");
			break;
		}
#if defined(ALLOW_PARANOID_CHECKING)
		if (nevents != 1) {
			fprintf(stderr, "readpacket exs_qdequeue: dequeued %d "
				"items, expected 1\n", nevents);
			break;
		}
		if (recv_vector[0].exs_evt_type != EXS_EVT_RECV) {
			fprintf(stderr, "readpacket exs_qdequeue: dequeued item"
				" type %d, expected %d\n",
				recv_vector[0].exs_evt_type,
				EXS_EVT_RECV);
			break;
		}
#endif
		TRACE_DBG(LOG_UTILITIES,
			"readpacket exs_qdequeue item type %d is RECV, errno %d",
			recv_vector[0].exs_evt_type,
			recv_vector[0].exs_evt_errno);
		if (recv_vector[0].exs_evt_errno != 0) {
			errno = recv_vector[0].exs_evt_errno;
			perror("readpacket exs_qdequeue status");
			break;
		}
		n = recv_vector[0].exs_evt_union.exs_evt_xfer.exs_evt_length;
		TRACE_DBG(LOG_UTILITIES,
			"readpacket exs_qdequeue item len %d", n);
		if (n == 0) {
			/* hit EOF */
			break;
		}
	} while (0);
	return n;		/* return total number of bytes actually read */
}	/* readpacket */


static int setup_attributes(int fd, char *message,
		const struct conn_params *params)
{
	long	old_value;

	if (params->small_packet_max_size != INT_MAX) {
		if ((old_value = exs_fcntl(fd, EXS_F_SETSPMAXSIZE,
					params->small_packet_max_size)) < 0) {
			fprintf(stderr, "%s exs_fcntl EXS_F_SETSPMAXSIZE: %s\n",
				message, strerror(errno));
			return -1;
		}
	}

	if (params->max_inline != INT_MAX) {
		if ((old_value = exs_fcntl(fd, EXS_F_SETINLINEMAXSIZE,
						params->max_inline)) < 0) {
			fprintf(stderr, "%s exs_fcntl EXS_F_SETINLINEMAXSIZE: %s\n",
				message, strerror(errno));
			return -1;
		}
	}

	if (params->comp_cpu != INT_MAX) {
		if ((old_value = exs_fcntl(fd, EXS_F_SETCOMPTHREADCPU,
						params->comp_cpu)) < 0) {
			fprintf(stderr, "%s exs_fcntl EXS_F_SETCOMPTHREADCPU: %s\n",
				message, strerror(errno));
			return -1;
		}
	}

	if (params->max_credits != INT_MAX) {
		if ((old_value = exs_fcntl(fd, EXS_F_SETFLOWCONTROLCREDITS,
						params->max_credits)) < 0) {
			fprintf(stderr, "%s exs_fcntl EXS_F_SETFLOWCONTROLCREDITS: %s\n",
				message, strerror(errno));
			return -1;
		}
	}

	if (params->stream_buf_size != INT_MAX) {
		if ((old_value = exs_fcntl(fd, EXS_F_SETSTREAMBUFSIZE,
						params->stream_buf_size)) < 0) {
			fprintf(stderr, "%s exs_fcntl EXS_F_SETSTREAMBUFSIZE: %s %ld\n",
				message, strerror(errno), old_value);
			return -1;
		}
	}

	if ((old_value = exs_fcntl(fd, EXS_F_SETFD, params->fd_flags)) < 0) {
		fprintf(stderr, "%s exs_fcntl EXS_F_SETFD %0x: %s\n",
			message, params->fd_flags, strerror(errno));
		return -1;
	}

	return 0;
}	/* setup_attributes */


/* expects message to end with colon (:) if user wants it printed */
int print_attributes(int fd, char *message)
{
	long	old_value;

	if ((old_value = exs_fcntl(fd, EXS_F_GETSPMAXSIZE, 0)) < 0){
		fprintf(stderr, "%s exs_fcntl EXS_F_GETSPMAXSIZE: %s\n",
			message, strerror(errno));
		return -1;
	}
	fprintf(stderr, "%s SMALL_PACKET_MAX_SIZE = %ld\n", message, old_value);

	errno = 0;
	if ((old_value = exs_fcntl(fd, EXS_F_GETCOMPTHREADCPU, 0)) < 0){
		if (errno != 0) {
			fprintf(stderr, "%s exs_fcntl EXS_F_GETCOMPTHREADCPU: %s\n",
				message, strerror(errno));
			return -1;
		}
	}
	fprintf(stderr, "%s COMP_THREAD_CPU = %ld\n", message, old_value);

	if ((old_value = exs_fcntl(fd, EXS_F_GETFLOWCONTROLCREDITS, 0)) < 0){
		fprintf(stderr, "%s exs_fcntl EXS_F_GETFLOWCONTROLCREDITS: %s\n",
			message, strerror(errno));
		return -1;
	}
	fprintf(stderr, "%s FLOW_CONTROL_CREDITS = %ld\n", message, old_value);

	if ((old_value = exs_fcntl(fd, EXS_F_GETSTREAMBUFSIZE, 0)) < 0){
		/* errno==EINVAL probably means that the fd is SOCK_SEQPACKET */
		if (errno != EINVAL) {
			fprintf(stderr, "%s exs_fcntl EXS_F_GETSTREAMBUFSIZE: %s\n",
				message, strerror(errno));
			return -1;
		}
	} else {
		fprintf(stderr, "%s STREAM_BUF_SIZE = %ld\n",
				message, old_value);
	}

	if ((old_value = exs_fcntl(fd, EXS_F_GETFD, 0)) < 0) {
		fprintf(stderr, "%s exs_fcntl EXS_F_GETFD: %s\n",
			message, strerror(errno));
		return -1;
	}

	fprintf(stderr, "%s EXS_FD_SP_COPY_UNREG_ON_SEND = %ld\n",
			message, (old_value & EXS_FD_SP_COPY_UNREG_ON_SEND));

	fprintf(stderr, "%s EXS_FD_SP_COPY_UNREG_ON_RECV = %ld\n",
			message, (old_value & EXS_FD_SP_COPY_UNREG_ON_RECV));

	fprintf(stderr, "%s EXS_FD_BYPASS_PARAM_CHECKING = %ld\n",
			message, (old_value & EXS_FD_BYPASS_PARAM_CHECKING));

	fprintf(stderr, "%s EXS_FD_NODELAY = %ld\n",
			message, (old_value & EXS_FD_NODELAY));

	fprintf(stderr, "%s EXS_FD_BUSYPOLL = %ld\n",
			message, (old_value & EXS_FD_BUSYPOLL));

	return 0;
}	/* print_attributes */


/* convert and print IP address as dotted-decimal or colon-hex string */
void
print_ip_address(struct sockaddr *addr, char *who, char *message)
{
	char	text_buf[TEXT_SIZE+1];

	if (addr->sa_family == PF_INET) {
		/* IPv4 address */
		if (inet_ntop(((struct sockaddr_in *)addr)->sin_family,
				&((struct sockaddr_in *)addr)->sin_addr,
				text_buf, TEXT_SIZE) == NULL) {
			fprintf(stderr, "%s %s %s: %s\n", message, who,
				"inet_ntop IPv4", strerror(errno));
		} else {
			fprintf(stderr, "%s %s at IPv4 address %s port %d\n",
				message, who, text_buf,
				ntohs(((struct sockaddr_in*)addr)->sin_port));
		}
	} else if (addr->sa_family == PF_INET6) {
		/* IPv6 address */
		if (inet_ntop(((struct sockaddr_in6 *)addr)->sin6_family,
				&((struct sockaddr_in6 *)addr)->sin6_addr,
				text_buf, TEXT_SIZE) == NULL) {
			fprintf(stderr, "%s %s %s: %s\n", message, who,
				"inet_ntop IPv6", strerror(errno));
		} else {
			fprintf(stderr, "%s %s at IPv6 address %s port %d\n",
				message, who, text_buf,
				ntohs(((struct sockaddr_in6*)addr)->sin6_port));
		}
	} else {
		fprintf(stderr, "%s %s %s: bad sa_family %d\n", message, who,
			"print_ip_address", addr->sa_family);
	}
}	/* print_ip_address */


/* Opens a new EXS socket to server on node server_node, port server_port.
 * Returns in server_addr the server's IPv4 address structure,
 * and in client_addr the client's IPv4 address structure.
 * Returns an fd that is connected to the server, or -1 on an error.
 */
int openclient(char *server_port, char *server_node,
		int socktype, struct sockaddr *server_addr,
		struct sockaddr *client_addr,
		const struct conn_params *params)
{
	int		fd, err, i, ret = -1;
	socklen_t	len;
	char		local_node[MAXHOSTNAMELEN];
	struct addrinfo	*aptr, hints;
	exs_qhandle_t	connect_queue;
	int		nevents;
	exs_event_t	connect_vector[CONNECT_ITEMS];


	/* get internet name of the local host node on which we are running */
	if (gethostname(local_node, MAXHOSTNAMELEN) < 0) {
		perror("openclient gethostname");
		goto err0;
	}

	/* set up the name of the remote host node for the server */
	if (server_node == NULL)
		server_node = local_node;	/* default to local node */

	/* give error only if all families fail in this loop */
	for (i = 0, err = -1; i < 2 && err != 0; i++) {
		/* get structure for remote host node on which server resides */
		memset(&hints, 0, sizeof(hints));
		hints.ai_family = family[i];
		hints.ai_socktype = SOCK_STREAM;
		err = getaddrinfo(server_node, server_port, &hints, &aptr);
		fprintf(stderr, "openclient %d getaddrinfo server_node %s, "
			"port %s, err %d\n", i, server_node, server_port, err);
	}
	if (err != 0) {
		/* all families failed, give error */
		fprintf(stderr, "openclient %s: %s port %s\n",
			server_node, server_port, gai_strerror(err));
		errno = ENOTCONN;	/* something non-zero */
		goto err0;
	}

	print_ip_address(aptr->ai_addr, "server address", "openclient");

	/* exs client needs to create an event_queue for connecting to server */
	connect_queue = exs_qcreate(CONNECT_DEPTH);
	if (connect_queue == EXS_QHANDLE_INVALID) {
		perror("openclient exs_qcreate");
		goto err1;
	}

	/* open an exs_socket for exs_client to connect on */
	if ((fd = exs_socket(aptr->ai_family, socktype, 0)) < 0) {
		perror("openclient exs_socket");
		goto err2;
	}

	if (setup_attributes(fd, "openclient", params)) {
		goto err2;
	}

	/* connect this socket to the server's Internet address */
	TRACE_DBG(LOG_UTILITIES,
		"openclient calling exs_connect, ai_family %d, ai_socktype %d",
		aptr->ai_family, aptr->ai_socktype);

	if (exs_connect(fd, aptr->ai_addr, aptr->ai_addrlen, 0,
			NULL, connect_queue, aptr) < 0 ) {
		perror("openclient exs_connect");
		goto err3;
	}

	TRACE_DBG(LOG_UTILITIES, "openclient back from exs_connect ok, calling exs_qdequeue, connect_vector %p",
			connect_vector);

	/* in exs need to wait for connect to complete asynchronously */
	nevents = exs_qdequeue(connect_queue,connect_vector,CONNECT_ITEMS,NULL);

	TRACE_DBG(LOG_UTILITIES,
		"openclient exs_qdequeue nevents=%d connect_vector[0] type %d is %s",
		nevents, connect_vector[0].exs_evt_type,
		connect_vector[0].exs_evt_type == EXS_EVT_CONNECT
				? "CONNECT" : "not CONNECT");
	if (nevents < 0) {
		perror("openclient exs_qdequeue");
		goto err3;
	}
#if defined(ALLOW_PARANOID_CHECKING)
	if (nevents != 1) {
		fprintf(stderr, "openclient exs_qdequeue: dequeued %d items, "
				" expected 1\n", nevents);
		goto err3;
	}
	if (connect_vector[0].exs_evt_type != EXS_EVT_CONNECT) {
		fprintf(stderr, "openclient exs_qdequeue: dequeued item"
			" type %d, expected %d\n",
			connect_vector[0].exs_evt_type,
			EXS_EVT_CONNECT);
		goto err3;
	}
#endif
	if (connect_vector[0].exs_evt_errno != 0) {
		errno = connect_vector[0].exs_evt_errno;
		perror("openclient exs_qdequeue status");
		goto err3;
	}

	if (client_addr) {
		/* caller wants local address assigned to this client */
		len = sizeof(struct sockaddr_in6);
		if (exs_getsockname(fd, client_addr, &len) < 0 ) {
			perror("openclient exs_getsockname");
			goto err3;
		}

	if (server_addr != NULL)
		/* caller wants server addess info returned */
		memcpy(server_addr, aptr->ai_addr, aptr->ai_addrlen);
	}

	/* we are now successfully connected to a remote server */
	ret = fd;		/* return fd of connected socket */
	goto err2;

err3:
	ret = errno;	/* save this across call to exs_blocking_close */
	if (exs_blocking_close(fd) < 0) {
		perror("openclient exs_blocking_close");
	}
	errno = ret;
	ret = -1;
err2:
	if  (exs_qdelete(connect_queue) < 0) {
		perror("openclient exs_qdelete(connect_queue)");
	}
err1:
	freeaddrinfo(aptr);
err0:
	return ret;
}	/* openclient */



#define LISTENING_DEPTH		3


/* Opens a new exs_socket for the exs_server at interface server_name,
 * port server_port.  Returns in server_addr the exs_server's internet
 * address structure.  The exs_server is NOT connected to a client on return.
 * Returns an fd that is a "listening post" on which to wait for connections,
 * and a new accept_queue on which to accept them.
 */
int
openserver(char *server_port, char *server_name, int listening_depth,
		int socktype, struct sockaddr *server_address,
		const struct conn_params *params)
{
	int		fd, err, i, ret = -1;
	socklen_t	len;
	char		local_node[MAXHOSTNAMELEN];
	struct addrinfo	*aptr, hints;
	char		*host_name, *host_port;
	

	/* get internet name of the local host node on which we are running */
	if (gethostname(local_node, MAXHOSTNAMELEN) < 0) {
		perror("openserver gethostname");
		goto err0;
	}

	if (server_name == NULL)
		host_name = "0.0.0.0";		/* all available interfaces */
	else
		host_name = server_name;

	if (server_port == NULL)
		host_port = "0";
	else
		host_port = server_port;

	/* give error only if all families fail in this loop */
	for (i = 0, err = -1; i < 2 && err != 0; i++) {
		/* get structure for local interface server is to use */
		memset(&hints, 0, sizeof(hints));
		hints.ai_family = family[i];
		hints.ai_socktype = SOCK_STREAM;
		hints.ai_flags = AI_PASSIVE;
		err = getaddrinfo(host_name, server_port, &hints, &aptr);
		/* err = -2 for "port Name or service not known"
		 * err = -9 for "port Address family for hostname not supported"
		 */
		fprintf(stderr, "openserver %d getaddrinfo host_name %s, "
			"port %s, err %d\n", i, host_name, host_port, err);
	}
	if (err != 0) {
		/* all families failed, give error */
		fprintf(stderr, "openserver %s: %s port %s\n",
			host_name, host_port, gai_strerror(err));
		errno = ENOTCONN;	/* something non-zero */
		goto err0;
	}

	print_ip_address(aptr->ai_addr, "server address", "openserver");

	/* open an exs_socket for exs_server to listen on */
	if ((fd = exs_socket(aptr->ai_family, socktype, 0)) < 0) {
		perror("openserver exs_socket");
		goto err1;
	}

	/* bind this exs_socket to the exs_server's Internet address */
	if (exs_bind(fd, aptr->ai_addr, aptr->ai_addrlen) < 0) {
		perror("openserver exs_bind");
		goto err2;
	}

	if (setup_attributes(fd, "openserver", params)) {
		goto err2;
	}

	/* set up listening backlog to accept connect requests from clients */
	if (listening_depth <= 0)
		listening_depth = LISTENING_DEPTH;
	if (exs_listen(fd, listening_depth) < 0) {
		perror("openserver exs_listen");
		goto err2;
	}

	if (server_address) {
		/* caller wants local address assigned to this exs_server */
		len = sizeof(struct sockaddr_in6);
		if (exs_getsockname(fd, server_address, &len) < 0 ){
			perror("openserver exs_getsockname");
			goto err2;
		}
	}

	/* we are now successfully established as an exs_server */
	ret = fd;
	goto err1;

err2:
	ret = errno;	/* save this across call to exs_blocking_close */
	if (exs_blocking_close(fd) < 0) {
		perror("openserver exs_blocking_close");
	}
	errno = ret;
	ret = -1;
err1:
	freeaddrinfo(aptr);
err0:
	return ret;			/* return fd of listening socket */
}	/* openserver */


void
fstatdump( int limit )
{
	int i;
	char *ptr;
	struct stat statbuf;
	char buffer[100];


	for (i = 0; i < limit; i++) {
		sprintf(buffer, "fd %d", i);
		if (fstat(i, &statbuf) < 0)
			perror(buffer);
		else {
			if (S_ISREG(statbuf.st_mode))
				ptr ="regular file";
			else if (S_ISDIR(statbuf.st_mode))
				ptr ="directory";
			else if (S_ISCHR(statbuf.st_mode))
				ptr ="character special file";
			else if (S_ISBLK(statbuf.st_mode))
				ptr ="block special file";
			else if (S_ISFIFO(statbuf.st_mode))
				ptr ="FIFO file";
			else if (S_ISLNK(statbuf.st_mode))
				ptr ="symbolic link";
			else if (S_ISSOCK(statbuf.st_mode))
				ptr ="socket";
			else
				ptr ="UNKNOWN FILE TYPE";
			printf("%s: %s\n", buffer, ptr);
		}
	}
}	/* fstatdump */

/* no_sigpipe - ignore sigpipe signals -- handle the error synchronously */
void
no_sigpipe(void)
{
	struct sigaction action;
	sigset_t	 pipe_mask;


	/* block the SIGPIPE signal */
	sigemptyset(&pipe_mask);
	sigaddset(&pipe_mask, SIGPIPE);
	if (sigprocmask(SIG_BLOCK, &pipe_mask, NULL) < 0) {
		perror("sigprocmask SIG_BLOCK SIGPIPE");
		exit(EXIT_FAILURE);
	}

	/* now set the action to ignore the SIGPIPE signal */
	action.sa_handler = SIG_IGN;
	sigemptyset(&action.sa_mask);
	action.sa_flags = 0;
	if (sigaction(SIGPIPE, &action, NULL)  < 0) {
		perror("sigaction SIGPIPE");
		exit(EXIT_FAILURE);
	}
}	/* no_sigpipe */

double
percentage(int secs, int usecs, int top_secs, int top_usecs)
{
	if (secs > 0 || usecs > 0) {
		return 100.0*((double)top_secs + ((double)top_usecs)/1000000.0)
		     / ((double)secs + ((double)usecs)/1000000.0);
	} else {
		return 0.0;
	}
}	/* percentage */

void
get_relative_time(struct timespec *start_time, int *secs, int *usecs,
							const char *message)
{
	struct timespec	stop_time;

	if (clock_gettime(CLOCK_EXS, &stop_time) < 0) {
		fprintf(stderr, "%s clock_gettime stop %s\n",
						message, strerror(errno));
		*secs = 0;
		*usecs = 0;
	} else {
		if (start_time->tv_nsec > stop_time.tv_nsec) {
			stop_time.tv_nsec += 1000000000;
			stop_time.tv_sec -= 1;
		}
		*secs = stop_time.tv_sec - start_time->tv_sec;
		*usecs = (stop_time.tv_nsec - start_time->tv_nsec) / 1000;
	}
}	/* get_relative_time */


void
get_relative_usage(struct rusage *start_usage, int *user_secs, int *user_usecs,
		int *kernel_secs, int *kernel_usecs, const char *message)
{
	struct rusage	stop_usage;

	if (getrusage(RUSAGE_SELF, &stop_usage) < 0) {
		fprintf(stderr, "%s getrusage stop %s\n",
						message, strerror(errno));
		*user_secs = 0;
		*user_usecs = 0;
		*kernel_secs = 0;
		*kernel_usecs = 0;
	} else {
		/* first compute user-space time for this process */
		if (start_usage->ru_utime.tv_usec
						> stop_usage.ru_utime.tv_usec) {
			stop_usage.ru_utime.tv_usec += 1000000;
			stop_usage.ru_utime.tv_sec -= 1;
		}
		*user_secs = stop_usage.ru_utime.tv_sec
					- start_usage->ru_utime.tv_sec;
		*user_usecs = stop_usage.ru_utime.tv_usec
					- start_usage->ru_utime.tv_usec;
		/* next compute kernel-space time for this process */
		if (start_usage->ru_stime.tv_usec
						> stop_usage.ru_stime.tv_usec) {
			stop_usage.ru_stime.tv_usec += 1000000;
			stop_usage.ru_stime.tv_sec -= 1;
		}
		*kernel_secs = stop_usage.ru_stime.tv_sec
					- start_usage->ru_stime.tv_sec;
		*kernel_usecs = stop_usage.ru_stime.tv_usec
					- start_usage->ru_stime.tv_usec;
	}
}	/* get_relative_usage */

/* vi: set autoindent tabstop=8 shiftwidth=8 : */
